#include <stdexcept>
#include "object.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Object::Object(Layout &l):
	layout(l),
	shape(0)
{ }

Snap Object::get_snap_node(unsigned) const
{
	throw out_of_range("Object::get_snap_node");
}

bool Object::snap(Snap &sn, float limit, SnapType what) const
{
	if(what&SNAP_NODE)
	{
		unsigned nsn = get_n_snap_nodes();
		for(unsigned i=0; i<nsn; ++i)
		{
			Snap node = get_snap_node(i);
			Vector span = sn.position-node.position;
			if(dot(span, span)<limit*limit)
			{
				sn = node;
				return true;
			}
		}
	}

	return false;
}

bool Object::snap_to(const Object &other, float limit, SnapType what)
{
	if(what==SNAP_DEFAULT)
		what = get_default_snap_type_to(other);

	if(!what)
		return false;

	unsigned nsn = get_n_snap_nodes();
	for(unsigned i=0; i<nsn; ++i)
	{
		Snap sn = get_snap_node(i);
		Snap ssn = sn;
		if(other.snap(ssn, limit, what))
		{
			if(what==SNAP_NODE)
				set_rotation(rotation+ssn.rotation-sn.rotation-Angle::half_turn());
			else
			{
				Angle adiff = wrap_balanced((ssn.rotation-sn.rotation)*2.0f)/2.0f;
				set_rotation(rotation+adiff);
			}
			sn = get_snap_node(i);
			set_position(position+ssn.position-sn.position);
			return true;
		}
	}

	return false;
}

Object *Object::get_link(unsigned) const
{
	throw out_of_range("Object::get_link");
}

bool Object::break_link(Object &other)
{
	unsigned nls = get_n_link_slots();
	for(unsigned i=0; i<nls; ++i)
		if(get_link(i)==&other)
			return break_link(i);

	return false;
}

void Object::break_links()
{
	unsigned nls = get_n_link_slots();
	for(unsigned i=0; i<nls; ++i)
		break_link(i);
}

const Shape *Object::get_shape() const
{
	if(shape)
		return shape;
	return get_type().get_shape();
}

bool Object::collide_ray(const Ray &ray, float *d) const
{
	const Shape *s = get_shape();
	if(!s)
		return false;

	Transform reverse_trans = Transform::rotation(rotation, Vector(0, 0, -1))*
		Transform::translation(-position);
	if(d)
	{
		vector<Geometry::SurfacePoint<float, 3> > points = s->get_intersections(reverse_trans.transform(ray));
		if(points.empty())
			return false;
		*d = points.front().distance;
		return true;
	}
	else
		return s->check_intersection(reverse_trans.transform(ray));
}

BoundingBox Object::get_bounding_box() const
{
	const Shape *s = get_shape();
	if(!s)
		return BoundingBox();

	Transform trans = Transform::translation(position)*
		Transform::rotation(rotation, Vector(0, 0, 1));
	return trans.transform(s->get_axis_aligned_bounding_box());
}

} // namespace R2C2
